from urlparse import urlparse 

import sys
import lxml.etree as et
import logging
import urllib
import time
import json
import cStringIO
import re
import xml.sax.saxutils as su

import splunk
import splunk.rest as rest
import splunk.entity as entity
import splunk.auth
import splunk.models.dashboard as sm_dashboard
import splunk.models.dashboard_panel as sm_dashboard_panel
import splunk.models.saved_search as sm_saved_search
import splunk.search
import splunk.search.searchUtils
from splunk.util import normalizeBoolean

import splunk.pdf.pdfgen_views as pv
import splunk.pdf.pdfgen_utils as pu
import splunk.pdf.pdfgen_chart as pc

import splunk.pdf.pdfrenderer as pdfrenderer

logger = pu.getLogger()

ERROR_MSG = _("Unable to render PDF.")
DEFAULT_FEED_COUNT = 500
DEFAULT_TIMEOUT = 3600 # in seconds
DEFAULT_MAX_ROWS_PER_TABLE = 1000
DEFAULT_INCLUDE_SPLUNK_LOGO = True
DEFAULT_PAPER_SIZE =   'letter'
DEFAULT_PAPER_ORIENTATION = 'portrait'

class ArgError(Exception):
    def __init__(self, message):
        self.message = message
    def __str__(self):
        return repr(self.value)

class TimeoutError(Exception):
    def __str__(self):
        return repr(self.value)

class PDFGenHandler(splunk.rest.BaseRestHandler):
    _queryArgs = {}
    _deletePdfFileAfterSettingResponse = True
    _timeoutTime = None

    _views = []
    _inputSids = {}
    
    _title = "Untitled"
    _dashboardName = None
    _reportName = None
    _viewType = None
    _paperSize = DEFAULT_PAPER_SIZE
    _namespace = None
    _owner = None
    _timeoutDuration = DEFAULT_TIMEOUT
    _maxRowsPerTable = DEFAULT_MAX_ROWS_PER_TABLE
    _includeSplunkLogo = DEFAULT_INCLUDE_SPLUNK_LOGO
    _cidFontList = None
    _now = None
    _timestampStr = ""

    _touchSearchJobsLastTime = None
    _pdfBuffer = None
    _pdfRenderer = None

    VIEW_TYPE_DASHBOARD = 'dashboard'
    VIEW_TYPE_REPORT =    'report'

    ARG_INPUT_DASHBOARD = 'input-dashboard'
    ARG_INPUT_REPORT = 'input-report'
    ARG_INPUT_PAPERSIZE = 'paper-size'
    ARG_INPUT_SID = 'sid'
    ARG_INPUT_NAMESPACE = 'namespace'
    ARG_INPUT_OWNER = 'owner'
    ARG_INPUT_TIMEOUT = 'timeout'
    ARG_INPUT_MAX_ROWS_PER_TABLE = 'max-rows-per-table'
    ARG_INPUT_INCLUDE_SPLUNK_LOGO = 'include-splunk-logo'
    ARG_INPUT_NOW = 'now'
    
    _validQueryArgs = [
        ARG_INPUT_DASHBOARD,
        ARG_INPUT_REPORT,
        ARG_INPUT_PAPERSIZE,
        ARG_INPUT_SID,
        ARG_INPUT_NAMESPACE,
        ARG_INPUT_TIMEOUT,
        ARG_INPUT_MAX_ROWS_PER_TABLE,
        ARG_INPUT_INCLUDE_SPLUNK_LOGO,
        ARG_INPUT_NOW
        ]
    
    SID_VALIDATOR_STR="sid_([0-9]+)"
    sidRE=re.compile(SID_VALIDATOR_STR)

    POLLING_WAIT_TIME = 0.5
    TOUCH_SEARCH_INTERVAL = 10

    ALERT_ACTIONS_ENTITY = '/configs/conf-alert_actions'
    LIMITS_ENTITY = '/configs/conf-limits'

    def handle_GET(self):
        logger.debug("pdfgen/render request: " + str(self.request))

        if not self._initialize():
            return

        if not self._render():
            return

        self._respond()

    def _initialize(self):
        try:
            self._initParameters()
            
            if self._viewType == self.VIEW_TYPE_DASHBOARD:
                (self._title, self._views) = pv.getDashboardTitleAndPanels(dashboardName=self._dashboardName, namespace=self._namespace, owner=self._owner, sessionKey=self.sessionKey)
            elif self._viewType == self.VIEW_TYPE_REPORT:
                self._views.append(pv.Report(savedSearchName=self._reportName, namespace=self._namespace, owner=self._owner, sessionKey=self.sessionKey))
                self._title = self._reportName
            
            self._handlePresetSearchIDs()
            
            # instantiate the pdfRenderer object with a file-like object
            self._pdfBuffer = cStringIO.StringIO()
            self._pdfRenderer = pdfrenderer.PDFRenderer(title=self._title, outputFile=self._pdfBuffer, paperSize=self._paperSize, timestamp=self._timestampStr, includeSplunkLogo=self._includeSplunkLogo, cidFontList=self._cidFontList)
            return True
        except ArgError as e:
            self.response.setStatus(400)
            self.response.write(e.message)
        except Exception as e:
            errorMsg = "Exception raised while preparing to render \"%s\" to PDF. %s" % (self._title, e)
            logger.error(errorMsg)
            self._outputError([errorMsg])
        return False 

    def _render(self):
        # start the first 3 searches
        self._inactiveViews = self._views[:]        
        for i in range(3):
            self._startNextSearch()

        # iterate over views and render each to the pdf
        viewPrepErrors = []
        for i, view in enumerate(self._views):
            try:
                if view.requiresSearchJobObj() and view.getSearchJobObj() == None:
                    raise Exception("No search job available for view that requires search job")

                self._keepAllSearchesAlive()
                self._checkForTimeout()
               
                isLastView = (i == len(self._views) - 1) 
                self._renderView(self._pdfRenderer, view, lastView = isLastView)
                
                self._startNextSearch()

            except TimeoutError:
                errorMsg = "Timeout while trying to prepare \"%s\" for rendering to PDF." % self._title
                self._outputTimeoutError(errorMsg)
                return
            except Exception as e:
                errorMsg = "Exception raised while trying to prepare \"%s\" for rendering to PDF. %s" % (self._title, e)
                logger.error(errorMsg)
                viewPrepErrors.append(errorMsg)

        # if we weren't able to render any views successfully, output an error as response to endpoint
        if len(viewPrepErrors) == len(self._views):
            self._outputError(viewPrepErrors)
            logger.error("No views prepared without exceptions. Bailing out of PDF render.")
            return False
            
        try:
            self._pdfRenderer.save()
        except Exception as e:
            errorMsg = "Exception raised while trying to render \"%s\" to PDF. %s" % (self._title, e)
            logger.error(errorMsg)
            self._outputError([errorMsg])
            return False

        return True

    def _respond(self):
        # save and write out the file
        try:
            self.response.write(self._pdfBuffer.getvalue())
            self.response.setHeader('content-type', 'application/pdf')
            # override normal cache-control header to fix problem on ie6-ie8 (see SPL-50739)
            self.response.setHeader('cache-control', 'max-age=0, must-revalidate')
            self.response.setStatus(200)
            self._pdfBuffer.close()
        except Exception as e:
            errorMsg = "Exception raised while trying to respond. Rendering \"%s\" to PDF. %s" % (self._title, e)
            logger.error(errorMsg)
            self._outputError([errorMsg])
            return False
        return True

    def _outputTimeoutError(self, message):
        self.response.write("PDF Generation timed out: %s" % message)
        self.response.setHeader('content-type', 'text/html')
        self.response.setStatus(408)

    def _outputError(self, errorDetailsArray):
        outputErrorMsg = ERROR_MSG + "<br/><ul>"
        for errorDetail in errorDetailsArray:
            outputErrorMsg = outputErrorMsg + "<li>" + su.escape(errorDetail) + "</li>"
        outputErrorMsg = outputErrorMsg + "</ul>"
        self.response.write(outputErrorMsg)
        self.response.setHeader('content-type', 'text/html')
        self.response.setStatus(400)

    def _initParameters(self):
        self._owner=self.request['userName']
        self._initQueryArgs()
        self._initLimitsDefaults()       
        self._initAlertActionsDefaults()
 
        # initialize view type
        if self._hasQueryArg(self.ARG_INPUT_DASHBOARD):
            self._dashboardName = self._getQueryArg(self.ARG_INPUT_DASHBOARD, "No input dashboard specified")
            self._viewType = self.VIEW_TYPE_DASHBOARD
        elif self._hasQueryArg(self.ARG_INPUT_REPORT):
            self._reportName = self._getQueryArg(self.ARG_INPUT_REPORT, "No input report specified")
            self._viewType = self.VIEW_TYPE_REPORT

        # initialize papersize
        if self._hasQueryArg(self.ARG_INPUT_PAPERSIZE):
            paperSizeArg = self._getQueryArg(self.ARG_INPUT_PAPERSIZE, "No input paper size specified").lower()
            if paperSizeArg in pdfrenderer.PAPERSIZES:
                self._paperSize = paperSizeArg
            else:
                raise ArgError("Paper size " + paperSizeArg + " not valid")
        logger.debug("pdf-init paper-size=%s" % self._paperSize)

        # initialize include-splunk-logo
        self._includeSplunkLogo = normalizeBoolean(self._queryArgs.get(self.ARG_INPUT_INCLUDE_SPLUNK_LOGO, self._includeSplunkLogo))
        logger.debug("pdf-init include-splunk-logo=%s" % self._includeSplunkLogo)

        # initialize max-row-per-table
        if self._hasQueryArg(self.ARG_INPUT_MAX_ROWS_PER_TABLE):
            maxRowsPerTableArg = self._getQueryArg(self.ARG_INPUT_MAX_ROWS_PER_TABLE, "No max rows per table specified")
            try:
                self._maxRowsPerTable = int(maxRowsPerTableArg)
            except:
                raise ArgError("max-rows-per-table=%s is invalid, must be an integer" % maxRowsPerTableArg)
        logger.debug("pdf-init max-rows-per-table=%s" % (str(self._maxRowsPerTable)))

        # initialize timeout
        if self._hasQueryArg(self.ARG_INPUT_TIMEOUT):
            self._timeoutDuration = int(self._getQueryArg(self.ARG_INPUT_TIMEOUT, "No timeout specified"))
        logger.debug("pdf-init timeoutDuration=%s" % self._timeoutDuration)
        self._startTimeoutClock()

        # initialize time of report
        self._initTimeOfReport()

        # check for SIDs
        if self._viewType is self.VIEW_TYPE_REPORT:
            if self._hasQueryArg(self.ARG_INPUT_SID):
                self._inputSids[0] = (self._getQueryArg(self.ARG_INPUT_SID, "No input sid specified"))
        else:
            for argK, argV in self._queryArgs.items():
                if self.ARG_INPUT_SID in argK:
                    # we want the panel sequence number which is retrieved from "sid_<seqNum>"
                    match = self.sidRE.match(argK)
                    if match != None and len(match.groups(0)) > 0:
                        seqNum = match.groups(0)[0]
                        if len(seqNum) > 0:
                            self._inputSids[int(seqNum)] = argV
                            logger.debug("sid seqNum=%s value=%s" % (seqNum, argV))

        # get namespace/owner
        self._namespace = self._getQueryArgOrDefault(self.ARG_INPUT_NAMESPACE, None)
        if self._hasQueryArg(self.ARG_INPUT_OWNER):
            self._owner = self._getQueryArg(self.ARG_INPUT_OWNER, "No input owner specified")

        self._validateParameters()

        self._timestampStr = splunk.search.searchUtils.getFormattedTimeForUser(
            sessionKey=self.sessionKey,
            user = self._owner,
            namespace = self._namespace,
            now = self._now
            )

    def _initTimeOfReport(self):
        # 1) try to use get time-of-report from the query string
        # 2) if not possible, get current UTC epoch time
        if self._hasQueryArg(self.ARG_INPUT_NOW):
            nowStr = self._getQueryArg(self.ARG_INPUT_NOW, "No now argument specified")
            try:
                self._now = float(nowStr)
            except:
                logger.warning("invalid now argument passed to pdfgen/render arg=%s" % nowStr)
        if self._now is None:
            self._now = time.time() 

        logger.info("pdf time-of-report=%s" % self._now)

    def _validateParameters(self):
        # raise ArgError if there are problems with any parameters
        if self._dashboardName is None and self._reportName is None:
            raise ArgError("PDF endpoint must be called with one of the following query args: 'input-dashboard=<dashboard-id>' or 'input-report=<report-id>'")

    def _initQueryArgs(self):
        self._queryArgs = self.request['query']

        for arg in self._queryArgs:
            logger.debug("Testing argument arg=%s value=%s" % (arg, self._queryArgs[arg]))
            if arg not in self._validQueryArgs:
                # check for sid_N for dashboards
                if self.sidRE.match(arg) is None:            
                    logger.warn("Invalid argument passed to pdfgen/render. arg=%s value=%s" % (arg, self._queryArgs[arg]))

    def _hasQueryArg(self, argKey):
        return argKey in self._queryArgs

    def _getQueryArgOrDefault(self, argKey, default):
        if self._hasQueryArg(argKey):
            return self._queryArgs[argKey]
        else:
            return default

    def _getQueryArg(self, argKey, errorMessage):
        if self._hasQueryArg(argKey):
            return self._queryArgs[argKey]
        else:
            raise ArgError(errorMessage)

    def _initAlertActionsDefaults(self):
        """ use alertActions entity to determine default papersize
            return in form of "<size>" or "<size>-landscape"
        """
        paperSize = DEFAULT_PAPER_SIZE
        paperOrientation = DEFAULT_PAPER_ORIENTATION
        try:
            settings = entity.getEntity(self.ALERT_ACTIONS_ENTITY, 'email', sessionKey=self.sessionKey)
            # paperSize is 'letter', 'legal', 'A4', etc
            paperSize = settings.get('reportPaperSize') or DEFAULT_PAPER_SIZE
            # paperOrientation is 'portrait' or 'landscape'
            paperOrientation = settings.get('reportPaperOrientation') or DEFAULT_PAPER_ORIENTATION
            self._includeSplunkLogo = normalizeBoolean(settings.get('reportIncludeSplunkLogo', self._includeSplunkLogo))
            cidFontListString = settings.get('reportCIDFontList', '') or ''
            self._cidFontList = cidFontListString.split(' ')
        except Exception as e:
            logger.error("Could not access or parse email stanza of alert_actions.conf. Error=%s" % str(e))

        if paperOrientation == 'landscape':
            self._paperSize = paperSize + '-landscape'
        else:    
            self._paperSize = paperSize


    def _initLimitsDefaults(self):
        """ use limits entity to determine defaults
        """
        try:
            settings = entity.getEntity(self.LIMITS_ENTITY, 'pdf', sessionKey=self.sessionKey)
            logger.debug("limitsPDFStanza=%s" % (str(settings)))
            self._maxRowsPerTable = int(settings.get('max_rows_per_table', self._maxRowsPerTable))
            self._timeoutDuration = int(settings.get('render_endpoint_timeout', self._timeoutDuration))
        except Exception as e:
            logger.error("Could not access or parse pdf stanza of limits.conf. Error=%s" % str(e))
 
    #
    # handling searches
    #

    def _handlePresetSearchIDs(self):
        if len(self._inputSids) > 0:
            for view in self._views:
                index = view.getViewIndex()
                if index in self._inputSids.keys():
                    try:
                        searchJobObj = splunk.search.SearchJob(self._inputSids[index], hostPath=None, sessionKey=self.sessionKey)
                    except Exception as e:
                        logger.warning("Nonexistent search job object for sid: " + self._inputSids[index])
                        searchJobObj = None

                    if searchJobObj != None:
                        if searchJobObj.isExpired():
                            logger.warning("Expired search job object for sid: " + self._inputSids[index])
                        else:
                            view.setSearchJobObj(searchJobObj)

    def _startNextSearch(self):
        if len(self._inactiveViews) == 0:
            return

        view = self._inactiveViews.pop(0)
        if view.getSearchJobObj() == None:
            # start the search!
            logger.debug("dispatching search view: %s" % view.getTitle())
            if not view.dispatchSearch(overrideNowTime=self._now):
                logger.error("could not dispatch search view: %s" % view.getTitle())
    
    def _keepAllSearchesAlive(self):
        currentTime = time.time()
        logger.debug("currentTime = %s self._touchSearchJobsLastTime = %s" % (currentTime, self._touchSearchJobsLastTime))
        if self._touchSearchJobsLastTime != None:
            delta = currentTime - self._touchSearchJobsLastTime             
            if delta < self.TOUCH_SEARCH_INTERVAL:
                return

        self._touchSearchJobsLastTime = currentTime
        for view in self._views:
            view.touchSearchJob()

    #
    # Timeout
    #

    def _startTimeoutClock(self):
        currentTime = int(time.time())
        self._timeoutTime = currentTime + self._timeoutDuration 
        logger.debug("PDF timeout setup: currentTime=%s timeoutDuration=%s timeoutTime=%s" % (currentTime, self._timeoutDuration, self._timeoutTime))

    def _checkForTimeout(self):
        currentTime = int(time.time())
        timedOut = currentTime > self._timeoutTime
        if timedOut:
            logger.info("PDF timeout. timeoutDuration = %s timeoutTime = %s currentTime = %s" % (self._timeoutDuration, self._timeoutTime, currentTime))
            for view in self._views:
                view.cancelSearch()
            raise TimeoutError()         

    #
    # Rendering
    #

    def _renderView(self, pdfRenderer, view, lastView=False):
        """ render an individual panel """
       
        if view.requiresSearchJobObj(): 
            while not view.isSearchComplete() and not view.isRealtime:
                time.sleep(self.POLLING_WAIT_TIME)
                
                self._keepAllSearchesAlive()
                self._checkForTimeout()

        pdfRenderer.conditionalPageBreak()
        self._renderViewHeader(pdfRenderer, view)

        types = view.getRenderTypes()

        for type in types:
            if type is 'chart':
                self._renderChart(pdfRenderer, view)
            elif type is 'table':
                self._renderTable(pdfRenderer, view)
            elif type is 'event':
                self._renderEvents(pdfRenderer, view)
            elif type is 'single':
                self._renderSingle(pdfRenderer, view)
            elif type is 'list':
                self._renderList(pdfRenderer, view)
            elif type is 'html':
                self._renderHtml(pdfRenderer, view)
            else:
                pdfRenderer.renderText("No render option for type = " + type)
                logger.warning("PDFGenHandler::_renderView> No render option for type = " + type)

        if not lastView:
            pdfRenderer.spaceBetween()

    def _renderViewHeader(self, pdfRenderer, view):
        title = view.getTitle()
        if title != None and len(title) > 0:
            pdfRenderer.renderText(title, style = pdfRenderer.TITLE_STYLE)
        description = view.getDescription()
        if description != None and len(description) > 0:
            pdfRenderer.renderText(description)

    def _renderTable(self, pdfRenderer, view):
        """ render a table of results """
        # get options
        options = view.getOptions()
        displayRowNumbers = False
        if "displayRowNumbers" in options:
            displayRowNumbers = normalizeBoolean(options['displayRowNumbers'])

        # build 2-d list of lists
        data = []
        resultsExist = False

        # get results object
        results = view.getSearchJobResults()
        
        # determine field set
        explicitFieldList = view.getSearchParams().get(pu.SP_FIELD_LIST, [])
        fieldOrder = []
        if len(explicitFieldList) > 0:
            for field in explicitFieldList:
                if field in results.fieldOrder and field not in fieldOrder:
                    fieldOrder.append(field)
            if len(fieldOrder) == 0:
                logger.warning("%s: results.fieldOrder does not contain any explicitly specified fields: %s" % (view.getTitle(), explicitFieldList))
                return 
        else:
            fieldOrder = self._renderTable_restrictFields(results.fieldOrder)
        
        if len(fieldOrder) == 0:
            pdfRenderer.renderText("No results found.")
            return

        results = view.getSearchJobResults()
        for i, result in enumerate(results):
            if i >= self._maxRowsPerTable:
                break

            if (i > 0) and ((i % 100) == 0):
                self._keepAllSearchesAlive()    

            resultsExist = True

            # not every result row in the results list will contain a cell for every column in the table
            # fill in missing cells with the empty string
            values = []
            for field in fieldOrder:
                if field in result.fields:
                    fieldValues = result.get(field, None)
                    logger.debug("fieldValues=%s len(fieldValues)=%s isinstance(fieldValues, splunk.search.RawEvent)=%s$" % (fieldValues, len(fieldValues), isinstance(fieldValues, splunk.search.RawEvent)))
                    fieldValuesStr = ""
                    if isinstance(fieldValues, splunk.search.RawEvent):
                        fieldValuesStr = fieldValues
                    elif len(fieldValues) > 1:
                        fieldValueStrings = [str(x) for x in fieldValues]
                        if "##__SPARKLINE__##".startswith(fieldValueStrings[0]):
                            fieldValuesStr = ','.join(fieldValueStrings) 
                        else:
                            fieldValuesStr = '\n'.join(fieldValueStrings)
                        logger.debug("fieldValueStrings=%s fieldValuesStr=%s" % (fieldValueStrings, fieldValuesStr))
                    else:
                        fieldValuesStr = fieldValues[0]
                    values.append(fieldValuesStr)
                else:
                    values.append("")

            data.append(values)

        columnWidths = []
        if fieldOrder[0] == "_time":
            columnWidths.append(1.33 * pdfRenderer.ONE_INCH)

        if resultsExist:
            pdfRenderer.renderTable(data, headerRow = fieldOrder, columnWidths=columnWidths, displayLineNumbers=displayRowNumbers)
        else:
            logger.warning("PDFGenHandler::_renderTable> no results for table")
            pdfRenderer.renderText("No results found.")

    def _renderTable_restrictFields(self, inputFieldOrder):
        """ restrict the fields that are output in tables
            do not display any fields that start with an underscore
            except for _raw and _time
            position _time, if present, at the front of fieldOrder, and
            position _raw, if present, at the end of fieldOrder
            ---
            for the time being, if _time and _raw are both present,
            do not show any other field
        """
        timeFieldPresent = "_time" in inputFieldOrder
        rawFieldPresent = "_raw" in inputFieldOrder

        # do not include any fields that start with '_'
        publicFields = []
        for field in inputFieldOrder:
            if len(field) > 0 and field[0] != '_':
                publicFields.append(field)

        # build up output fieldOrder
        fieldOrder = []

        # always show _time first
        if timeFieldPresent:
            fieldOrder.append("_time")

        # if both _time and _raw present, do not show other fields
        if timeFieldPresent == False or rawFieldPresent == False:
            for field in publicFields:
                fieldOrder.append(field)

        # always show _raw last
        if rawFieldPresent:
            fieldOrder.append("_raw")

        logger.debug("PDFGenHandler::_renderTable_restrictFields> inputFieldOrder=" + str(inputFieldOrder) + " fieldOrder=" + str(fieldOrder))

        return fieldOrder

    def _renderList(self, pdfRenderer, view):
        """ render a list display """

        # build 2-d list of lists
        data = []
        resultsExist = False

        results = view.getSearchJobResults()
        fieldOrder = results.fieldOrder

        # get options
        optionsDict = view.getOptions()
        labelField = optionsDict.get("labelField", None)
        valueField = optionsDict.get("valueField", None)
        initialSort = optionsDict.get("initialSort", labelField)
        initialSortDir = optionsDict.get("initialSortDir", "asc")

        # validate options
        if labelField is None or valueField is None:
            logger.warning("PDFGenHandler::_renderList> missing either labelField or valueField in list")
            return

        # build up 2-d data set
        data = []
        for result in results:
            label = str(result.get(labelField, default=""))
            value = str(result.get(valueField, default=""))

            logger.debug("PDFGenHandler::_renderList> label, value: " + label + ", " + value)

            if label != "" or value != "":
                row = [label, value]
                data.append(row)

        # sort by initialSort and initialSortDir
        sortIndex = 0
        if initialSort == valueField:
            sortIndex = 1
        sortDescending = initialSortDir.lower() == "desc"

        data = sorted(data, key=lambda row: row[sortIndex], reverse=sortDescending)
        logger.debug("PDFGenHandler::_renderList> data: " +str(data))

        # render the data as lines of text
        if len(data) > 0:
            for row in data:
                pdfRenderer.renderText(row[0] + " " + row[1])
        else:
            pdfRenderer.renderText("No results for list")


    def _renderSingle(self, pdfRenderer, view):
        """ render a SingleValue display """
        resultsExist = False
        singleValue = ""
        classValue = "none"
        classStyle = None
        optionsDict = view.getOptions()

        results = view.getSearchJobResults()
        try:
            result = results[0]
            resultsExist = True
            if "field" in optionsDict:
                singleValue = result.fields.data[optionsDict["field"]]
            else:
                singleValue = result.fields.data.values()[0]

            if "classField" in optionsDict:
                classValue = str(result.fields.data[optionsDict["classField"]])

            if classValue in pdfRenderer.singleValueStyles:
                logger.debug("PDFGenHandler::_renderSingle> pdfRenderer.singleValueStyles[classValue]: " + str(pdfRenderer.singleValueStyles[classValue]))
                classStyle = pdfRenderer.singleValueStyles[classValue]
            logger.debug("PDFGenHandler::_renderSingle> singleValueStyles: " + str(pdfRenderer.singleValueStyles))
            logger.debug("PDFGenHandler::_renderSingle> classValue: " + str(type(classValue)) + ":" + str(classValue) + ", classStyle: " + str(classStyle))

        except IndexError:
            resultsExist = False

        if resultsExist:
            optionsDict = view.getOptions()
            beforeLabel = ""
            afterLabel = ""
            if "beforeLabel" in optionsDict:
                beforeLabel = optionsDict["beforeLabel"] + " "
            if "afterLabel" in optionsDict:
                afterLabel = " " + optionsDict["afterLabel"]
            pdfRenderer.renderText(beforeLabel + str(singleValue) + afterLabel, style=classStyle)

    def _renderChart(self, pdfRenderer, view):
        """ render a chart from the results """
        # set the props
        props = {"exportMode":"true","enableChartClick":"false","enableLegendClick":"false"}
        props = dict(props.items() + view.getChartProps().items())

        # set the fetch option output_mode to 'json_cols'
        #  this will cause the data returned by getFeed to be
        #  pre-populated into the columnar form that the charting code wants
        view.getSearchJobObj().setFetchOptions(output_mode="json_cols", time_format="")

        feedCount = DEFAULT_FEED_COUNT
        if 'data.count' in view.getChartProps():
            feedCount = view.getChartProps()['data.count']

        feed = view.getSearchJobFeed(feedCount=feedCount)
        logger.debug("_renderChart> feed: " + str(feed) + ", " + str(type(feed)))
        
        if feed is None or len(feed.strip()) == 0:
            logger.warning("_renderChart> feed: " + str(feed) + ", " + str(type(feed)))

            pdfRenderer.renderText("No results found.")
            return

        feed = json.loads(feed)

        # extract columns and fields from the search results
        data = feed['columns']
        fields = feed['fields']

        if len(data) == 0:
            pdfRenderer.renderText("No results found.")
            return            

        # set up the chart object
        chart = pc.Chart(data, fields, props)

        # build the chart
        buildSuccess = chart.build()
        if buildSuccess:
            svgString = chart.getSvg()
            pdfRenderer.renderSvgString(svgString, title=None)
        else:
            logger.error("renderChart chart building error")


    def _renderEvents(self, pdfRenderer, view):
        """ render a listing of events """
        # get options -- TODO: should be refactored with _renderTable (and beyond: should have a generalized system for getting options)
        options = view.getOptions()
        displayRowNumbers = False
        if "displayRowNumbers" in options:
            displayRowNumbers = normalizeBoolean(options['displayRowNumbers'])

        data = []
        events = view.getSearchJobEvents()
        for i, event in enumerate(events):
            if i >= self._maxRowsPerTable:
                break
    
            if (i > 0) and ((i % 100) == 0):
                self._keepAllSearchesAlive()    

            data.append([str(event.time), event.raw])

        if len(data) == 0:
            pdfRenderer.renderText("No matching events found.")
        else:
            # only set column widths for the line number and timestamp columns, setting None
            # for the last column provides for automatically setting its width
            colWidths = [pdfRenderer.ONE_INCH * 1.3, None]

            # we want to hard wrap the raw event column only
            colHardWraps = [False, True]

            pdfRenderer.renderTable(data, columnWidths=colWidths, columnHardWraps=colHardWraps, columnVAlignments=['TOP','TOP'], displayLineNumbers=displayRowNumbers)


    def _renderHtml(self, pdfRenderer, view):
        title = view.getTitle()
        if title is None:
            title = "Untitled panel"

        options = view.getOptions()
        logger.debug("PDFGenHandler::_renderHtml> options: " + str(options))
        if pu.PP_RAW_CONTENT not in options or options[pu.PP_RAW_CONTENT] is None or len(options[pu.PP_RAW_CONTENT]) == 0:
            logger.warning("PDFGenHandler::_renderHtml> rawContent key not in optionsDict for view: " + title)
            return

        html = et.HTML(options[pu.PP_RAW_CONTENT])

        for text in html.itertext():
            pdfRenderer.renderText(text)



