import cgi
import datetime
import json
import logging
import operator
import os
import tempfile
import time
import traceback

import cherrypy

import splunk, splunk.auth, splunk.util, splunk.entity, splunk.rest as rest
from splunk.appserver.mrsparkle import *
from splunk.appserver.mrsparkle.lib import jsonresponse, util, cached
from splunk.models.license import License, Pool, Stack, Slave, Message, SelfConfig, Group
from splunk.util import normalizeBoolean
from splunk.models.inputs import SplunkTCPInput
from splunk.models.clustering import ClusterMasterPeer, ClusterMasterBucket, ClusterMasterInfo, ClusterSlaveBucket, ClusterSlaveInfo, ClusterConfig, ClusterSlaveControl, ClusterSearchheadInfo, ClusterMasterGeneration
import splunk.models.search_bar as SearchBar
from splunk.models.server_config import SplunkdConfig
import splunk.entity as en

#TODO: remove this
from splunk.appserver.mrsparkle.lib.msg_pool import MsgPoolMgr, UI_MSG_POOL

logger = logging.getLogger('splunk.appserver.controllers.clustering')
#logger.setLevel(logging.DEBUG)

# define default stack name
DEFAULT_STACK_NAME = 'enterprise'

# define set of license groups that allow pool creation
POOLABLE_GROUPS = ['Enterprise']

# define the model value for a pool's catch-all slave list
CATCHALL_SLAVE_LIST = ['*']


class ClusteringException(Exception):
    """ Used to signal invalid inputs """
    pass 




class ClusteringController(BaseController):
    """
    Clustering
    """

    #
    # attach common template args
    #

    def render_template(self, template_path, template_args = {}):
        template_args['appList'] = self.get_app_manifest()
        return super(ClusteringController, self).render_template(template_path, template_args)
    
        
    def get_app_manifest(self):
        '''
        Returns a dict of all available apps to current user
        '''
        
        output = cached.getEntities('apps/local', search=['disabled=false','visible=true'], count=-1)
                
        return output

    #
    # Clustering Dashboard
    #

    @route('/')
    @expose_page(methods=['GET'])
    def show_landing_page(self, **unused):
        '''
        Determines which landing page to show, depending on whether the instance is a master or slave
        '''

        logger.info("\n\n\n In show_landing_page:\n\n\n ");  

        clusterStatus = self.getClusterConfig()

        logger.info("mode: " + clusterStatus.mode)
        if clusterStatus.mode == 'master':
            self.redirect_to_url(
                    ['manager','system','clustering','dashboard'], 
                    _qs={
                    }
            )
        elif clusterStatus.mode == 'slave':
            splunkd_config = SplunkdConfig.get()
            slave_guid = splunkd_config.guid
            self.redirect_to_url(
                    ['manager','system','clustering','slave','peers', slave_guid], 
                    _qs={
                    }
            )
        elif clusterStatus.mode == 'searchhead':
            splunkd_config = SplunkdConfig.get()
            self.redirect_to_url(
                    ['manager','system','clustering','searchhead','details'], 
                    _qs={
                    }
            )
        else:
            isCancel = unused.get('cancel', None)

            if isCancel: 
                # User hit cancel button on configuration page and clustering is disabled.  Redirect to the manager page.  
                self.redirect_to_url(
                    ['manager'], 
                    _qs={
                    }
                )
            else: 
                self.redirect_to_url(
                    ['manager','system','clustering','configureClustering'], 
                    _qs={
                    }
                )

    def mode_to_string(self, mode):
        if not mode:
            return 'disabled'
        if mode == 'eTypeMaster':
            mode = 'Master'
        elif mode == 'eTypeSlave': 
            mode = 'Slave'

        return mode

    def show_error(self, msg):
        #msgid = MsgPoolMgr.get_poolmgr_instance()[UI_MSG_POOL].push('error', msg)
        #e = Exception(msg)
        #cherrypy.session['clusterException'] = e

        self.redirect_to_url(
                ['manager','system','clustering','configureClustering'] 
        )

    #
    # Clustering Dashboard (NEW)
    #

    @route('/:path=dashboard')
    @expose_page(methods=['GET'])
    def show_dashboard(self, **kwargs):
        '''
        Renders the clustering dashboard page
        '''


        logger.debug("\n\n\n In show_dashboard: \n\n\n") 

        clusterStatus = self.getClusterConfig()
        if clusterStatus.mode != 'master':
            raise cherrypy.HTTPError(403)         


        peer_count = 0
        peer_statuses = {}
        peers = {}

        try: 
            generation = ClusterMasterGeneration.all()[0]
        except IndexError: 
            self.redirect_to_url(
                ['manager','system','clustering','configureClustering'] 
            )

 
        gen_peers = generation.generation_peers

        #TODO: only count a peer as searchable/up if it's known generation is the current generation.

        for peer in ClusterMasterPeer.all():
            peers[peer.name] = {'up':(peer.status != 'Down' and peer.status != "Restarting" and peer.status != "GracefulShutdown" and peer.status != "ShuttingDown" and peer.status != "Pending") and peer.name in gen_peers,
                                'label':peer.label, 
                                'name' : peer.name,
                                'status' : peer.status,
                                'buckets':peer.buckets}

            if peer.status == 'Up' and peer.name not in gen_peers:
                peer.status = 'Unsearchable'
            

            peer_statuses[peer.status] = peer_statuses.get(peer.status, 0) + 1
            peer_count += 1



        indexes = {}
        for bucket in ClusterMasterBucket.all(count_per_req=10000):
            if bucket.standalone:
                continue

            if bucket.frozen:
                continue

            parts = bucket.name.split('~')

            #if parts[2] not in peers:
            #    peers[parts[2]] = {'up':False,
            #                       'label':''}

            index_data = indexes.get(parts[0], {})

            # sometimes we don't get a size: SPL-54219
            bucket_size = bucket.bucket_size or 1

            index_data['buckets'] = index_data.get('buckets', 0) + 1
            index_data['size'] = index_data.get('size', 0) + bucket_size
      
            searchable_copies = 0
            indexed_copies = 0
            replicated_copies = 0
    
    
            for (peer,state) in bucket.peers.items():
                if not peers[peer]['up']:
                    continue

                replicated_copies += 1

                if state['search_state'] != 'Searchable':
                    continue

                indexed_copies += 1

                if state['bucket_flags'] == '0x0':
                    continue

                searchable_copies += 1

            if searchable_copies > 0:
                index_data['searchable_size'] = index_data.get('searchable_size', 0) + bucket_size

            if indexed_copies > 0:
                index_data['index_up_size'] = index_data.get('index_up_size', 0) + bucket_size

            if indexed_copies > 1:
                index_data['indexed_size'] = index_data.get('indexed_size', 0) + bucket_size

            if replicated_copies > 1:
                index_data['replica_up_size'] = index_data.get('replica_up_size', 0) + bucket_size

            if replicated_copies > 2:
                index_data['replicated_size'] = index_data.get('replicated_size', 0) + bucket_size

            index_data['min_copies'] = min(replicated_copies, index_data.get('min_copies', replicated_copies))
            index_data['min_searchable_copies'] = min(indexed_copies, index_data.get('min_searchable_copies', indexed_copies))
            searchable_size = index_data.get('searchable_size',0)
            #if searchable_size == index_data['size'] : 
            #    statusMessage = "OK"  
            #else: 
            #    statusMessage = "ERROR *****************************"
            #logger.debug("Examining bucket %s: searchable_size=%s, index_size=%s - %s" % (bucket.name, searchable_size, index_data['size'], statusMessage))
            indexes[parts[0]] = index_data

       
        # Determine global searchable health of the cluster

        isGenerationCommitFailure = generation.pending_last_reason is not None
        isSearchFactorMet = True
        isReplicationFactorMet = True
        isClusterSearchable = True 
        numSearchableIndexes = 0
        numUnsearchableIndexes = 0

        if isGenerationCommitFailure: 
            isClusterSearchable = False

        for (index, index_data) in indexes.items():
            searchable_size = index_data.get('searchable_size',0)
            if searchable_size >= index_data['size'] and searchable_size > 0: 
                index_data['isSearchable'] = "Yes"
                numSearchableIndexes += 1
            else: 
                logger.debug("Index %s is NOT searchable.  searchable_size=%s, index_size=%s" % (index, searchable_size, index_data['size'])) 
                index_data['isSearchable'] = "No"
                numUnsearchableIndexes += 1
                isClusterSearchable = False

            searchable_copies = index_data.get('min_searchable_copies',0)
            if searchable_copies == 0:
                isClusterSearchable = False
            if searchable_copies < clusterStatus.search_factor: 
                isSearchFactorMet = False      

            replicated_copies = index_data.get('min_copies',0)
            if replicated_copies == 0:
                isClusterSearchable = False
            if replicated_copies < clusterStatus.replication_factor: 
                isReplicationFactorMet = False      


            

        isPeerDown = True if peer_statuses.get("Up",0) < len(peers) else False

       
        # Fetch master information
        try: 
            masterInfo = ClusterMasterInfo.all()[0]
        except IndexError: 
            msg = "There was an error retrieving the master information.  Try restarting splunkd.  If the error persists, clear the [clustering] stanza in server.conf and restart splunkd." 
            self.show_error(msg)
     
        splunkd_config = SplunkdConfig.get()
        master_guid = splunkd_config.guid
        master_uri = self.splunkd_urlhost
      
         
        template_args = {
            'peers': peers,
            'peer_count' : peer_count,
            'peer_statuses' : peer_statuses,
            'isSearchFactorMet' : isSearchFactorMet,
            'isReplicationFactorMet' : isReplicationFactorMet,
            'isClusterSearchable' : isClusterSearchable,
            'isGenerationCommitFailure' : isGenerationCommitFailure,
            'isPeerDown' : isPeerDown,
            'numSearchableIndexes' : numSearchableIndexes,
            'numUnsearchableIndexes' : numUnsearchableIndexes,
            'clusterStatus' : clusterStatus,            
            'master_guid' : master_guid, 
            'master_uri' : master_uri, 
            'master_info' : masterInfo,
            'indexes' : indexes,
            'mode' : self.mode_to_string(clusterStatus.mode), 
        }


        return self.render_template('/clustering/clusteringDashboard.html', template_args)
        #return self.render_template('/clustering/master/dashboard.html', template_args)
 

    #
    # Clustering Dashboard
    #

    @route('/:path=dashboardold')
    @expose_page(methods=['GET'])
    def show_dashboard_old(self, **kwargs):
        '''
        Renders the clustering dashboard page
        '''


        logger.debug("\n\n\n In show_dashboard: \n\n\n") 

        clusterStatus = self.getClusterConfig()
        if clusterStatus.mode != 'master':
            raise cherrypy.HTTPError(403)         

        # Fetch the list of peers
        entities = ClusterMasterPeer.all()
        peers = SearchBar.filterEntities(kwargs, entities)
        SearchBar.initialize(kwargs)

        # Fetch master information
        try: 
            masterInfo = ClusterMasterInfo.all()[0]
        except IndexError: 
            msg = "There was an error retrieving the master information.  Try restarting splunkd.  If the error persists, clear the [clustering] stanza in server.conf and restart splunkd." 
            self.show_error(msg)
        
        # Fetch searchhead information
        try: 
            searchheadInfo = ClusterSearchheadInfo.all()[0]
        except IndexError: 
            msg = "There was an error retrieving the searchhead information.  Try restarting splunkd.  If the error persists, clear the [clustering] stanza in server.conf and restart splunkd." 
            self.show_error(msg)
            raise cherrypy.HTTPRedirect(self.make_url(['manager','system','clustering','configureClustering'], _qs=dict()), 303)
        except KeyError: 
            # master is not configured to be a searchhead
            searchheadInfo = None

        splunkd_config = SplunkdConfig.get()
        master_guid = splunkd_config.guid
        master_uri = self.splunkd_urlhost
      
        #Removing search head table from master dashboard page
        searchheadInfo = None
        
        template_args = {
            'peers': peers,
            'master' : clusterStatus,            
            'kwargs' : kwargs,
            'master_guid' : master_guid, 
            'master_uri' : master_uri, 
            'master_info' : masterInfo,
            'mode' : self.mode_to_string(clusterStatus.mode), 
            'searchhead_info' : searchheadInfo,
        }


        return self.render_template('/clustering/clusteringDashboard.html.old', template_args)
        #return self.render_template('/clustering/master/dashboard.html', template_args)
 
    #
    # Searchhead details page (searchhead's view)
    #

    @route('/:mode=searchhead/:path=details')
    @expose_page(methods=['GET'])
    def show_searchhead_details(self, **kwargs):
        '''
        Renders the details for a searchhead
        '''

        logger.debug("\n\n\n In show_searchhead_details: \n\n\n") 

        clusterStatus = self.getClusterConfig()
        
        if clusterStatus.mode != 'searchhead':
            raise cherrypy.HTTPError(403)         

        # Fetch searchhead information
        searchheadInfo = ClusterSearchheadInfo.all()
        if len(searchheadInfo) == 0: 
            msg = "There was an error retrieving the searchhead information.  Try restarting splunkd.  If the error persists, clear the [clustering] stanza in server.conf and restart splunkd." 
            self.show_error(msg)

        host_port_pair = self.splunkd_urlhost

        
        template_args = {
            'kwargs' : kwargs,
            'host_port_pair' : host_port_pair,
            'searchhead_info' : searchheadInfo,
            'mode' : self.mode_to_string(clusterStatus.mode).lower(), 
        }

        return self.render_template('/clustering/searchhead/details.html', template_args)



    #
    # Peer details page (slave's view)
    #

    @route('/:mode=slave/:path=peers/:peername')
    @expose_page(methods=['GET'])
    def show_peer_details_slave(self, peername=None, **kwargs):
        '''
        Renders the details for a given peer (slave's view)
        '''

        logger.debug("\n\n\n In show_peer_details_slave!!!!!!!!!!!!!!!!!!!!: \n\n\n") 

        clusterStatus = self.getClusterConfig()

        if clusterStatus.mode != 'slave':
            raise cherrypy.HTTPError(403)         

        # Fetch slave information
        try: 
            slaveInfo = ClusterSlaveInfo.all()[0]
        except IndexError: 
            msg = "There was an error retrieving the peer information.  Try restarting splunkd.  If the error persists, clear the [clustering] stanza in server.conf and restart splunkd." 
            self.show_error(msg)


        master_uri = clusterStatus.master_uri
        splunkd_config = SplunkdConfig.get()
        host_port_pair = self.splunkd_urlhost
        last_heartbeat = '?' 
        status = '?'  

        bucketEntities = ClusterSlaveBucket.all(count_per_req=10000).search('*')

        buckets = SearchBar.filterEntities(kwargs, bucketEntities)
        SearchBar.initialize(kwargs)
        peerLabel = SplunkdConfig.get().server_name

        template_args = {
            'kwargs' : kwargs,
            'peername' : peerLabel if peerLabel else peername,
            'status' : status,
            'config' : clusterStatus,            
            'master_uri' : master_uri,
            'host_port_pair' : host_port_pair,
            'last_heartbeat' : last_heartbeat,
            'buckets' : buckets,
            'slave_info' : slaveInfo, 
            'mode' : self.mode_to_string(clusterStatus.mode).lower(), 
        }

        return self.render_template('/clustering/slave/peerDetails.html', template_args)


    #
    # Peer details page (master's view)
    #

    @route('/:mode=master/:path=peers/:peername')
    @expose_page(methods=['GET','POST'])
    def show_peer_details_master(self, peername, controller_exception=None, action=None, **kwargs):
        '''
        Renders the details for a given peer (master's view)
        '''

        logger.debug("\n\n\n In show_peer_details_master: \n\n\n") 

        clusterStatus = self.getClusterConfig()

        if clusterStatus.mode != 'master':
            raise cherrypy.HTTPError(403)         
  

        # Get buckets associated with this peer
        bucketEntities = ClusterMasterBucket.all(count_per_req=10000).search(peername)
        buckets = SearchBar.filterEntities(kwargs, bucketEntities)
        finalBuckets = []
        for bucket in buckets: 
            peerAttributes = bucket.peers.get(peername,None)
            if not peerAttributes: 
                continue
            bucket.search_state = peerAttributes.get('search_state','*')
            bucket.status = peerAttributes.get('status','*')
            finalBuckets.append(bucket)
 
        # User is restarting peer 
        if cherrypy.request.method == 'POST' and action == 'restartPeer':
            uri = '/services/cluster/master/peers/%s/restart' % peername
            try: 
                result = en.controlEntity('enable', uri)
                logger.debug("result of restarting peer: %s " % result)
            except splunk.InternalServerError, e:
                logger.debug("Error occurred: %s" % e)
                controller_exception = e


        # Get peer metadata
        thePeer = None
        for peer in ClusterMasterPeer.all():
            if peer.name == peername:
                thePeer = peer
                break 

        master_uri = self.splunkd_urlhost
        host_port_pair = thePeer.host_port_pair
        last_heartbeat = thePeer.last_heartbeat
        rawdata_rcv_port = thePeer.replication_port
        status = thePeer.status

        SearchBar.initialize(kwargs)
    
        template_args = {
            'kwargs' : kwargs,
            'controller_exception' : controller_exception, 
            'peername' : thePeer.label if thePeer.label else thePeer.name,
            'peerGuid' : thePeer.name, 
            'status' : status,
            'master' : clusterStatus,            
            'master_uri' : master_uri,
            'host_port_pair' : host_port_pair,
            'last_heartbeat' : last_heartbeat,
            'rawdata_rcv_port' : rawdata_rcv_port,
            'buckets' : finalBuckets,
            'mode' : self.mode_to_string(clusterStatus.mode).lower(), 
        }


        return self.render_template('/clustering/master/peerDetails.html', template_args)
 
    #
    #  Decomission peer confirmation popup
    #

    @route('/:mode=master/:path=confirmDecommission')
    @expose_page(methods=['GET'])
    def decomission_peer_confim(self, peername, **kwargs):

        logger.debug("\n\n\n In decommission_peer_confirm: \n\n\n") 

        template_args = {'app':'search', 
                         'peername' : peername}
        return self.render_template('/clustering/master/decommissionPeerConfirm.html', template_args)

 
    #
    #  Decomission peer 
    #

    @route('/:mode=master/:path=startDecommission')
    @expose_page(must_login=True, trim_spaces=True, methods='POST')
    def decommission_peer(self, peername, **kwargs):

        logger.debug("\n\n\n In decommission_peer \n\n\n") 
        controller_exception = None
        clusterStatus = self.getClusterConfig()

        if clusterStatus.mode != 'master':
            raise cherrypy.HTTPError(403)         
  

        uri = '/services/cluster/master/peers/%s/decommission' % peername
        
        serverResponse, serverContent = rest.simpleRequest(uri, method='DELETE', raiseAllErrors=True)        

        if serverResponse.status != 200: 
            controller_exception = Exception('unhandled HTTP status=%s' % serverResponse.status)


        template_args = {
            'app' : 'search',
            'peername' : peername, 
            'controller_exception' : controller_exception
        }
        
        return self.render_template('/clustering/master/decommissionResult.html', template_args)
 





    @route('/:mode=slave/:path=buckets/:bucketname')
    @expose_page(methods=['GET'])
    def show_buckets_slave(self, bucketname=None, **unused): 

        clusterStatus = self.getClusterConfig()
        
        if clusterStatus.mode != 'slave': 
            raise cherrypy.HTTPError(403)         

        buckets = ClusterSlaveBucket.all(count_per_req=10000).search(bucketname) 
        logger.debug("buckets: %s, length: %s" % (buckets, len(buckets)))
        if len(buckets) == 0:
            return self.render_template('/clustering/slave/bucketDetails.html')

        theBucket = buckets[0]

        template_args = {
            'bucket': theBucket,
            'mode' : self.mode_to_string(clusterStatus.mode) 
        }    

        return self.render_template('/clustering/slave/bucketDetails.html', template_args)


    @route('/:mode=master/:path=allbuckets/')
    @expose_page(methods=['GET'])
    def show_all_buckets(self, **kwargs): 
        logger.debug("\n\n\nShowing all buckets.\n\n\n"); 

        clusterStatus = self.getClusterConfig()
        
        if clusterStatus.mode != 'master': 
            raise cherrypy.HTTPError(403)         

        # Fetch the list of buckets
        entities = ClusterMasterBucket.all(count_per_req=10000).search('*')
        buckets = SearchBar.filterEntities(kwargs, entities)
        SearchBar.initialize(kwargs)


        logger.debug("buckets: %s, length: %s" % (buckets, len(buckets)))

        template_args = {
            'buckets': buckets,
            'mode' : self.mode_to_string(clusterStatus.mode), 
            'kwargs' : kwargs,
        }    

        return self.render_template('/clustering/master/allBuckets.html', template_args)



    @route('/:mode=master/:path=buckets/:bucketname')
    @expose_page(methods=['GET'])
    def show_bucket_details_master(self, bucketname=None, **unused): 
        logger.debug("\n\n\nIn show_bucket_details_master\n\n\n") 

        clusterStatus = self.getClusterConfig()
        
        if clusterStatus.mode != 'master': 
            raise cherrypy.HTTPError(403)         

        theBucket = None
        logger.debug("bucketname: %s" % bucketname)
        buckets = ClusterMasterBucket.all(count_per_req=10000).search(bucketname) 
        logger.debug("buckets: %s, length: %s" % (buckets, len(buckets)))
        if len(buckets) == 0:
            return self.render_template('/clustering/master/bucketDetails.html')

        theBucket = buckets[0]

        template_args = {
            'bucket': theBucket,
            'mode' : self.mode_to_string(clusterStatus.mode) 
        }    

        return self.render_template('/clustering/master/bucketDetails.html', template_args)

    def getClusterConfig(self, fromConfigureClustering=False):
        try: 
            cfg = ClusterConfig.all()[0]
            return cfg
        except IndexError: 
            raise Exception("The splunkd daemon cannot be reached by splunkweb. Check that there are no blocked network ports or that splunkd is still running.")
        except splunk.LicenseRestriction, e:
            if not fromConfigureClustering:
                self.redirect_to_url(
                    ['manager','system','clustering','configureClustering'] 
                )
            else:
                raise e

      

    def isValidUnsignedInt(self, x):
        try: 
            return int(x) >= 1 
        except ValueError: 
            return False
    #
    # Configure clustering
    #

    @route('/:path=configureClustering')
    @expose_page(methods=['GET','POST'])
    def show_configure_clustering(self, controller_exception=None, master_secret=None, slave_secret=None, searchhead_secret=None, num_copies=None, num_searchable_copies=None, is_clustering_enabled=None, formset_heartbeat_timeout=None, formset_mode=None, formset_master_uri=None, searchhead_master_uri=None, splunk_tcp_port=None,  **params):
        '''
        Renders the configure clustering page
        '''

        logger.debug("\n\n\n In show_configure_clustering: \n\n\n") 

        # Check if any clustering exceptions exist, will indicate the splunk must be restarted.  Show once and then remove from session. 
        controller_exception = cherrypy.session.get('clusterException',None)
        cherrypy.session['clusterException'] = None    


        try:
            clusterStatus = self.getClusterConfig(fromConfigureClustering=True)
        except splunk.LicenseRestriction, e:
            template_args = {
                'license_restriction': True,
                'controller_exception': e
            }
            return self.render_template('/clustering/402.html', template_args)
            
        logger.debug("replication_port: %s" % clusterStatus.replication_port)

        if cherrypy.request.method == 'POST':
            try: 
                clusterStatus = self.getClusterConfig()
                clusterStatus.mode = None

                logger.debug("is_clustering_enabled: %s" % is_clustering_enabled)
                if not is_clustering_enabled: 
                    clusterStatus.mode = 'disabled'

                if formset_mode == 'master': 
                    clusterStatus.mode = 'master'            
                    if not self.isValidUnsignedInt(formset_heartbeat_timeout): 
                        raise ClusteringException("Heartbeat timeout must be a valid unsigned integer greater than 0.")
                    clusterStatus.heartbeat_timeout = formset_heartbeat_timeout            

                    if not self.isValidUnsignedInt(num_copies): 
                        raise ClusteringException("Number of bucket copies must be a valid unsigned integer greater than 0.")
                    clusterStatus.replication_factor = num_copies               

                    if not self.isValidUnsignedInt(num_searchable_copies): 
                        raise ClusteringException("Number of searchable copies must be a valid unsigned integer greater than 0.")
                    if num_searchable_copies > num_copies: 
                        raise ClusteringException("Number of searchable copies must be less than or equal to the number of bucket copies.")
                    clusterStatus.search_factor = num_searchable_copies 

                    if master_secret is not None: 
                        clusterStatus.secret = master_secret 
                elif formset_mode == "slave": 
                    clusterStatus.mode = 'slave'
                    if not formset_master_uri: 
                        raise ClusteringException("Must enter required field(s): Master URI")

                    clusterStatus.master_uri = formset_master_uri 
                
                    if slave_secret is not None: 
                        clusterStatus.secret = slave_secret 

                    if not splunk_tcp_port:
                        raise ClusteringException("Must enter required field(s): Replication port")

                    clusterStatus.replication_port = splunk_tcp_port
                elif formset_mode == "searchhead": 
                    clusterStatus.mode = 'searchhead'
                    clusterStatus.master_uri = searchhead_master_uri 
                    if not searchhead_master_uri: 
                        raise ClusteringException("Must enter required field(s): Master URI")

                    if searchhead_secret is not None: 
                        clusterStatus.secret = searchhead_secret 

                if not clusterStatus.mode:
                    raise ClusteringException("Must select master, slave, or searchhead")

                clusterStatus.save()         
                clusterStatus = self.getClusterConfig()
            except ClusteringException, e: 
                controller_exception = e 
            except splunk.BadRequest, e:
                controller_exception = e
            except splunk.InternalServerError, e:  
                controller_exception = e

        splunkd_config = SplunkdConfig.get()
        master_guid = splunkd_config.guid
        master_uri =  clusterStatus.master_uri if clusterStatus.master_uri is not None else ''
        ip =  clusterStatus.ip if clusterStatus.ip is not None else ''


        template_args = {
            'license_restriction': False,
            'controller_exception' : controller_exception,
            'clusterStatus' : clusterStatus,            
            'master_uri' : master_uri, 
            'ip' : ip, 
            'mode' : (self.mode_to_string(clusterStatus.mode)).lower(),  
            'is_clustering_enabled' : False if not clusterStatus.mode or clusterStatus.mode == "disabled"  else True,
            'master_guid' : master_guid, 
        }





        return self.render_template('/clustering/configureClustering.html', template_args)




    @route('/:mode=master/:path=configure')
    @expose_page(methods=['GET','POST'])
    def show_configure_master(self, num_copies=None, num_searchable_copies=None, heartbeat_timeout=None, formset_cxn_timeout=None, formset_discard_timeout=None, **unused):
        '''
        Renders the master configuration page
        '''


        logger.debug("\n\n\n In show_configure_master: \n\n\n") 
        numIndexers = len(ClusterMasterPeer.all())

        clusterStatus = self.getClusterConfig()

        if clusterStatus.mode != 'master':
            raise cherrypy.HTTPError(403)         

      
        if cherrypy.request.method == 'POST':
            logger.debug("In configureMaster POST: numCopies: " + num_copies + ", num_searchable_copies: " + num_searchable_copies + " ... \n\n"); 
            clusterStatus.replication_factor = num_copies
            clusterStatus.hearbeat_timeout = heartbeat_timeout    
            clusterStatus.discard_timeout = 1
            clusterStatus.save()
      
        splunkd_config = SplunkdConfig.get()
        master_guid = splunkd_config.guid
        master_uri = self.splunkd_urlhost
        replication_factor_options = [['1','1'],['2','2'],['3','3']] 

        template_args = {
            'numIndexers': numIndexers,
            'master' : clusterStatus,            
            'master_guid' : master_guid, 
            'master_uri' : master_uri, 
            'replication_factor_options' : replication_factor_options,
            'selected_replication_factor': str(clusterStatus.replication_factor),
            'mode' : self.mode_to_string(clusterStatus.mode) 
        }

        return self.render_template('/clustering/configureMaster.html', template_args)



    #
    # Configure Cluster Slave
    #

    @route('/:mode=slave/:path=configure')
    @expose_page(methods=['GET','POST'])
    def show_configure_slave(self, master_uri=None, **unused):
        '''
        Renders the slave configuration page
        '''


        logger.debug("\n\n\n In show_configure_slave: \n\n\n") 

        clusterStatus = self.getClusterConfig()
        
        if clusterStatus.mode != 'slave':
             logger.debug("Raising error")
             raise cherrypy.HTTPError(403)         

        if cherrypy.request.method == 'POST':
            logger.debug("In configureSlave POST: master_uri: " + master_uri + " ... \n\n"); 
            clusterStatus.master_uri = master_uri
            clusterStatus.save()
      
        template_args = {
            'slaveStatus' : clusterStatus,            
            'mode' : self.mode_to_string(clusterStatus.mode) 
        }

        return self.render_template('/clustering/configureSlave.html', template_args)



