import cherrypy
import xml.sax.saxutils as su
from splunk.appserver.mrsparkle import *
from splunk.appserver.mrsparkle.lib.util import make_url
import logging
import splunk.rest, splunk.rest.format
import json
import re
import urllib

logger = logging.getLogger('splunk.appserver.controllers.proxy')

PDFGEN_RENDER_TIMEOUT_IN_SECONDS = 3600 # same as cherrypy.tools.sessions.timeout
                                        # since this module is loaded before settings are
                                        # fully loaded, we can't use web.conf value here

_PROXY_WHITE_LIST = [{'endpoint': '(__raw/services/)?pdfgen/render', 'methods': ['GET'], 'timeout': PDFGEN_RENDER_TIMEOUT_IN_SECONDS},
                     {'endpoint': '(__raw/services/)?pdfgen/is_available', 'methods': ['GET']},
                     {'endpoint': 'data/inputs/win-perfmon', 'methods': ['POST', 'GET']},
                     {'endpoint': 'apps/apptemplates', 'methods': ['GET']},
                     {'endpoint': 'authorization/capabilities/[^/]*/', 'methods': ['GET']},
                     {'endpoint': 'licenser/slaves', 'methods': ['GET']},
                     {'endpoint': 'data/inputs/oneshot/[^/]*', 'methods': ['GET']},
                     {'endpoint': 'data/transforms/extractions/[^/]*', 'methods': ['GET', 'DELETE', 'POST']},
                     {'endpoint': 'server/info/[^/]*', 'methods': ['GET']},
                     {'endpoint': 'data/inputs/tcp/raw/[^/]*/connections', 'methods': ['GET']},
                     {'endpoint': 'data/inputs/win-perfmon/[^/]*', 'methods': ['GET', 'DELETE', 'POST']},
                     {'endpoint': 'data/inputs/monitor/[^/]*/members', 'methods': ['GET']},
                     {'endpoint': 'apps/apptemplates/[^/]*', 'methods': ['GET']},
                     {'endpoint': 'data/inputs/ad', 'methods': ['POST', 'GET']},
                     {'endpoint': 'storage/passwords/[^/]*', 'methods': ['GET', 'DELETE', 'POST']},
                     {'endpoint': 'search/jobs/[^/]*', 'methods': ['GET', 'DELETE']},
                     {'endpoint': 'data/lookup-table-files/[^/]*', 'methods': ['GET', 'DELETE', 'POST']},
                     {'endpoint': 'properties/[^/]*', 'methods': ['GET', 'POST']},
                     {'endpoint': 'server/settings/[^/]*', 'methods': ['GET', 'POST']},
                     {'endpoint': 'storage/passwords', 'methods': ['POST', 'GET']},
                     {'endpoint': 'alerts/fired_alerts', 'methods': ['GET']},
                     {'endpoint': 'search/fields', 'methods': ['GET']},
                     {'endpoint': 'scheduled/views/[^/]*/history', 'methods': ['GET']},
                     {'endpoint': 'licenser/messages', 'methods': ['GET']},
                     {'endpoint': 'receivers/simple', 'methods': ['POST']},
                     {'endpoint': 'properties/[^/]*/[^/]*/[^/]*', 'methods': ['GET', 'POST']},
                     {'endpoint': 'data/props/lookups', 'methods': ['POST', 'GET']},
                     {'endpoint': 'directory/[^/]*', 'methods': ['GET']},
                     {'endpoint': 'licenser/pools', 'methods': ['POST', 'GET']},
                     {'endpoint': 'data/commands/[^/]*', 'methods': ['GET']},
                     {'endpoint': 'configs/conf-[^/]*', 'methods': ['GET', 'POST']},
                     {'endpoint': 'apps/local/[^/]*/setup', 'methods': ['GET']},
                     {'endpoint': 'saved/searches', 'methods': ['POST', 'GET']},
                     {'endpoint': 'data/outputs/tcp/syslog/[^/]*', 'methods': ['GET', 'DELETE', 'POST']},
                     {'endpoint': 'search/tags/[^/]*', 'methods': ['GET', 'POST', 'DELETE']},
                     {'endpoint': 'authentication/auth-tokens', 'methods': ['POST', 'GET']},
                     {'endpoint': 'data/inputs/script/[^/]*', 'methods': ['GET', 'DELETE', 'POST']},
                     {'endpoint': 'data/inputs/monitor/[^/]*', 'methods': ['GET', 'DELETE', 'POST']},
                     {'endpoint': 'saved/searches/[^/]*/acknowledge', 'methods': ['POST']},
                     {'endpoint': 'search/jobs/[^/]*/events', 'methods': ['GET']},
                     {'endpoint': 'licenser/stacks', 'methods': ['GET']},
                     {'endpoint': 'apps/local', 'methods': ['POST', 'GET']},
                     {'endpoint': 'data/inputs/tcp/ssl', 'methods': ['GET']},
                     {'endpoint': 'data/inputs/script/restart', 'methods': ['POST']},
                     {'endpoint': 'data/props/fieldaliases', 'methods': ['POST', 'GET']},
                     {'endpoint': 'search/timeparser', 'methods': ['GET']},
                     {'endpoint': 'licenser/licenses', 'methods': ['POST', 'GET']},
                     {'endpoint': 'search/distributed/peers/[^/]*', 'methods': ['GET', 'DELETE', 'POST']},
                     {'endpoint': 'data/inputs/script', 'methods': ['POST', 'GET']},
                     {'endpoint': 'data/props/lookups/[^/]*', 'methods': ['GET', 'DELETE', 'POST']},
                     {'endpoint': 'data/outputs/tcp/server/[^/]*/allconnections', 'methods': ['GET']},
                     {'endpoint': 'data/outputs/tcp/group/[^/]*', 'methods': ['GET', 'DELETE', 'POST']},
                     {'endpoint': 'authorization/capabilities', 'methods': ['GET']},
                     {'endpoint': 'authorization/roles', 'methods': ['POST', 'GET']},
                     {'endpoint': 'authorization/roles/[^/]*', 'methods': ['GET', 'DELETE', 'POST']},
                     {'endpoint': 'data/inputs/udp', 'methods': ['POST', 'GET']},
                     {'endpoint': 'data/props/fieldaliases/[^/]*', 'methods': ['GET', 'DELETE', 'POST']},
                     {'endpoint': 'receivers/stream', 'methods': ['POST']},
                     {'endpoint': 'data/inputs/oneshot', 'methods': ['POST', 'GET']},
                     {'endpoint': 'search/tags', 'methods': ['GET']},
                     {'endpoint': 'apps/local/[^/]*/update', 'methods': ['GET']},
                     {'endpoint': 'data/outputs/tcp/group', 'methods': ['POST', 'GET']},
                     {'endpoint': 'authentication/current-context/[^/]*', 'methods': ['GET']},
                     {'endpoint': 'scheduled/views/[^/]*/dispatch', 'methods': ['POST']},
                     {'endpoint': 'search/jobs', 'methods': ['GET', 'POST']},
                     {'endpoint': 'authentication/users', 'methods': ['POST', 'GET']},
                     {'endpoint': 'messages', 'methods': ['POST', 'GET']},
                     {'endpoint': 'licenser/groups/[^/]*', 'methods': ['GET', 'POST']},
                     {'endpoint': 'scheduled/views/[^/]*/scheduled_times', 'methods': ['GET']},
                     {'endpoint': 'authentication/users/[^/]*', 'methods': ['GET', 'DELETE', 'POST']},
                     {'endpoint': 'licenser/pools/[^/]*', 'methods': ['GET', 'DELETE', 'POST']},
                     {'endpoint': 'data/props/sourcetype-rename/[^/]*', 'methods': ['GET', 'DELETE', 'POST']},
                     {'endpoint': 'search/parser', 'methods': ['GET']},
                     {'endpoint': 'data/commands', 'methods': ['GET']},
                     {'endpoint': 'auth/login', 'methods': ['POST']},
                     {'endpoint': 'saved/searches/[^/]*', 'methods': ['GET', 'DELETE', 'POST']},
                     {'endpoint': 'scheduled/views/[^/]*', 'methods': ['GET', 'DELETE', 'POST']},
                     {'endpoint': 'data/indexes', 'methods': ['POST', 'GET']},
                     {'endpoint': 'search/fields/[^/]*', 'methods': ['GET']},
                     {'endpoint': 'apps/local/[^/]*/package', 'methods': ['GET']},
                     {'endpoint': 'search/jobs/[^/]*/timeline', 'methods': ['GET']},
                     {'endpoint': 'licenser/stacks/[^/]*', 'methods': ['GET']},
                     {'endpoint': 'indexing/preview/[^/]*', 'methods': ['GET']},
                     {'endpoint': 'licenser/groups', 'methods': ['GET']},
                     {'endpoint': 'data/transforms/extractions', 'methods': ['POST', 'GET']},
                     {'endpoint': 'search/distributed/config/[^/]*', 'methods': ['GET', 'DELETE', 'POST']},
                     {'endpoint': 'authentication/httpauth-tokens', 'methods': ['GET']},
                     {'endpoint': 'data/inputs/registry/[^/]*', 'methods': ['GET', 'DELETE', 'POST']},
                     {'endpoint': 'properties', 'methods': ['GET', 'POST']},
                     {'endpoint': 'data/props/extractions/[^/]*', 'methods': ['GET', 'DELETE', 'POST']},
                     {'endpoint': 'search/jobs/[^/]*/summary', 'methods': ['GET', 'GET', 'GET', 'GET']},
                     {'endpoint': 'search/fields/[^/]*/tags', 'methods': ['GET', 'POST']},
                     {'endpoint': 'server/control', 'methods': ['GET']},
                     {'endpoint': 'data/inputs/udp/[^/]*', 'methods': ['GET', 'DELETE', 'POST']},
                     {'endpoint': 'data/props/sourcetype-rename', 'methods': ['POST', 'GET']},
                     {'endpoint': 'server/settings', 'methods': ['GET']},
                     {'endpoint': 'search/typeahead', 'methods': ['GET']},
                     {'endpoint': 'search/jobs/[^/]*/search.log', 'methods': ['GET']},
                     {'endpoint': 'indexing/preview', 'methods': ['GET', 'POST']},
                     {'endpoint': 'search/jobs/export', 'methods': ['GET']},
                     {'endpoint': 'data/inputs/ad/[^/]*', 'methods': ['GET', 'DELETE', 'POST']},
                     {'endpoint': 'data/inputs/udp/[^/]*/connections', 'methods': ['GET']},
                     {'endpoint': 'search/jobs/[^/]*/control', 'methods': ['POST']},
                     {'endpoint': 'apps/appinstall', 'methods': ['POST']},
                     {'endpoint': 'data/inputs/win-event-log-collections', 'methods': ['POST', 'GET']},
                     {'endpoint': 'data/inputs/win-wmi-collections/[^/]*', 'methods': ['GET', 'DELETE', 'POST']},
                     {'endpoint': 'licenser/messages/[^/]*', 'methods': ['GET']},
                     {'endpoint': 'data/inputs/tcp/cooked', 'methods': ['POST', 'GET']},
                     {'endpoint': 'data/transforms/lookups', 'methods': ['POST', 'GET']},
                     {'endpoint': 'server/logger/[^/]*', 'methods': ['GET', 'POST']},
                     {'endpoint': 'search/jobs/[^/]*/results', 'methods': ['GET']},
                     {'endpoint': 'configs/conf-[^/]*/[^/]*', 'methods': ['GET', 'POST', 'DELETE']},
                     {'endpoint': 'server/control/restart', 'methods': ['POST']},
                     {'endpoint': 'data/inputs/win-wmi-collections', 'methods': ['POST', 'GET']},
                     {'endpoint': 'data/inputs/tcp/ssl/[^/]*', 'methods': ['GET', 'POST']},
                     {'endpoint': 'apps/local/[^/]*', 'methods': ['GET', 'DELETE', 'POST']},
                     {'endpoint': 'data/outputs/tcp/syslog', 'methods': ['POST', 'GET']},
                     {'endpoint': 'saved/searches/[^/]*/dispatch', 'methods': ['POST']},
                     {'endpoint': 'saved/searches/[^/]*/history', 'methods': ['GET']},
                     {'endpoint': 'saved/searches/[^/]*/suppress', 'methods': ['GET']},
                     {'endpoint': 'server/logger', 'methods': ['GET']},
                     {'endpoint': 'data/inputs/registry', 'methods': ['POST', 'GET']},
                     {'endpoint': 'data/indexes/[^/]*', 'methods': ['GET', 'POST']},
                     {'endpoint': 'data/indexes/[^/]*/enable', 'methods': ['POST']},
                     {'endpoint': 'data/indexes/[^/]*/disable', 'methods': ['POST']},
                     {'endpoint': 'saved/eventtypes/[^/]*', 'methods': ['GET', 'DELETE', 'POST']},
                     {'endpoint': 'data/inputs/tcp/cooked/[^/]*/connections', 'methods': ['GET']},
                     {'endpoint': 'server/info', 'methods': ['GET']},
                     {'endpoint': 'licenser/slaves/[^/]*', 'methods': ['GET']},
                     {'endpoint': 'data/inputs/tcp/raw', 'methods': ['POST', 'GET']},
                     {'endpoint': 'search/distributed/peers', 'methods': ['POST', 'GET']},
                     {'endpoint': 'authentication/current-context', 'methods': ['GET']},
                     {'endpoint': 'search/distributed/config', 'methods': ['GET']},
                     {'endpoint': 'data/inputs/tcp/raw/[^/]*', 'methods': ['GET', 'DELETE', 'POST']},
                     {'endpoint': 'data/outputs/tcp/default', 'methods': ['POST', 'GET']},
                     {'endpoint': 'authentication/httpauth-tokens/[^/]*', 'methods': ['GET', 'DELETE']},
                     {'endpoint': 'data/lookup-table-files', 'methods': ['POST', 'GET']},
                     {'endpoint': 'data/outputs/tcp/server', 'methods': ['POST', 'GET']},
                     {'endpoint': 'search/jobs/[^/]*/results_preview', 'methods': ['GET']},
                     {'endpoint': 'data/props/extractions', 'methods': ['POST', 'GET']},
                     {'endpoint': 'data/inputs/win-event-log-collections/[^/]*', 'methods': ['GET', 'DELETE', 'POST']},
                     {'endpoint': 'data/inputs/monitor', 'methods': ['POST', 'GET']},
                     {'endpoint': 'data/inputs/tcp/cooked/[^/]*', 'methods': ['GET', 'DELETE', 'POST']},
                     {'endpoint': 'data/transforms/lookups/[^/]*', 'methods': ['GET', 'DELETE', 'POST']},
                     {'endpoint': 'data/outputs/tcp/server/[^/]*', 'methods': ['GET', 'DELETE', 'POST']},
                     {'endpoint': 'data/outputs/tcp/default/[^/]*', 'methods': ['GET', 'DELETE', 'POST']},
                     {'endpoint': 'messages/[^/]*', 'methods': ['GET', 'DELETE']},
                     {'endpoint': 'licenser/licenses/[^/]*', 'methods': ['GET', 'DELETE']},
                     {'endpoint': 'alerts/fired_alerts/[^/]*', 'methods': ['GET', 'DELETE']},
                     {'endpoint': 'properties/[^/]*/[^/]*', 'methods': ['GET', 'POST']},
                     {'endpoint': 'saved/eventtypes', 'methods': ['POST', 'GET']},
                     {'endpoint': 'directory', 'methods': ['GET']},
                     {'endpoint': 'scheduled/views', 'methods': ['GET']},
                     {'endpoint': 'saved/searches/[^/]*/scheduled_times', 'methods': ['GET']},
                     {'endpoint': 'data/ui/views', 'methods': ['GET', 'POST']},
                     {'endpoint': 'data/ui/views/[^/]*', 'methods': ['GET', 'POST', 'DELETE']},]

def precompile_whitelist():
    for props in _PROXY_WHITE_LIST:
        regex_string = '(^|^services/|^servicesns/[^/]+/[^/]+/)%s$' % props['endpoint']
        regex = re.compile(regex_string)
        props['endpoint'] = regex
        
precompile_whitelist()
from datetime import datetime

class ProxyController(BaseController):
    """/splunkd"""

    @route('/*_proxy_path')
    @expose_page(must_login=False, methods=['GET', 'POST', 'DELETE'])
    def index(self, **args):
        
        if cherrypy.request.method in ['POST', 'DELETE'] and not cherrypy.config.get('enable_proxy_write'):
            return self.generateError(405, _('Write access to the proxy endpoint is disabled.'))
 
        sessionKey = cherrypy.session.get('sessionKey')    
        cherrypy.session.release_lock()

        if not sessionKey:
            logger.warn('proxy accessed without stored session key')
        
        # We have to handle the fact that CherryPy is going to %-decode
        # the URL, including any "/" (%2F). As such, we use the relative_uri
        # (which doesn't %-decode %2F), and simply re-encode that URL 
        relative_uri = cherrypy.request.relative_uri
        relative_uri = relative_uri[relative_uri.find("/splunkd")+9:]
        query_start = relative_uri.find("?")
        if query_start > -1:
            relative_uri = relative_uri[:query_start]

	uri = urllib.quote(relative_uri)
        
        if uri.startswith('__raw/'):
            # Don't parse any response even if it's a 404 etc
            rawResult = True
            uri = uri[6:]
        else:
            rawResult = False
            
        endpointProps = self.getAllowedEndpointProps(uri, cherrypy.request.method)
        if endpointProps is None:
            # endpoint not allowed
            logger.info("Resource not found: %s" % uri)
            raise cherrypy.HTTPError(404, _('Resource not found: %s' % uri))

        # Force URI to be relative so an attacker can't hit any arbitrary URL
        uri = '/' + uri

        if cherrypy.request.query_string:
            uri += '?' + cherrypy.request.query_string
            
        postargs = None
        body = None
        if cherrypy.request.method == 'POST':
            content_type = cherrypy.request.headers.get('Content-Type', '')
            if not content_type or content_type.find('application/x-www-form-urlencoded') > -1:
                # We use the body_params to avoid mixing up GET/POST arguments,
                # which is the norm with output_mode=json in Ace.
                postargs = cherrypy.request.body_params
            else:
                # special handing for application/json POST
                # cherrypy gives file descriptor for POST's
                body = cherrypy.request.body.read()
            
        proxyMode = False
        if 'authtoken' in args:
            proxyMode = True
       
        simpleRequestTimeout = splunk.rest.SPLUNKD_CONNECTION_TIMEOUT
        if 'timeout' in endpointProps:
                simpleRequestTimeout = endpointProps['timeout']
 
        try:
            serverResponse, serverContent = splunk.rest.simpleRequest(
                make_url(uri, translate=False, relative=True), 
                sessionKey,
                postargs=postargs,
                method=cherrypy.request.method,
                raiseAllErrors=True,
                proxyMode=proxyMode,
                rawResult=rawResult,
                jsonargs=body,
                timeout=simpleRequestTimeout
            )

            for header in serverResponse:
                cherrypy.response.headers[header] = serverResponse[header]
                
            if serverResponse.messages:
                return self.generateError(serverResponse.status, serverResponse.messages)
            
            if rawResult:
                cherrypy.response.status = serverResponse.status
            return serverContent
                
        except splunk.RESTException, e:
            logger.exception(e)
            return self.generateError(e.statusCode, e.extendedMessages)
            
        except Exception, e:
            logger.exception(e)
            return self.generateError(500, su.escape(str(e)))
            
            
    def getAllowedEndpointProps(self, uri, method):
        '''verify that that a given uri and associated method is white listed to be proxied to the endpoint.'''
       
        uri = uri.lower()
        for props in _PROXY_WHITE_LIST:
            if props['endpoint'].match(uri):
                if method in props['methods']:
                    return props 
        else:
            return None

    def generateError(self, status, messages=None):
        output = [splunk.rest.format.XML_MANIFEST, '<response>']
        output.append('<meta http-equiv="status" content="%s" />' % su.escape(str(status)))
        if messages:
            output.append('<messages>')
            
            if isinstance(messages, list):
                for msg in messages:
                    output.append('<msg type="%s">%s</msg>' % (su.escape(msg['type']), su.escape(msg['text'])))
            else:
                output.append('<msg type="ERROR">%s</msg>' % str(messages))
            output.append('</messages>')
            
        output.append('</response>')
        
        return '\n'.join(output)
        


